//
//  FontX.java
//
//  Copyright (c) 1996, Microsoft Corporation
//
// notes:
// 1.	cant override getFont.
// 2.	going to have to eventually override ALL methods of java.awt.Font,
//		leave place holders in code now.
//
package com.ms.awt;
import java.awt.Font;
import java.awt.Component;
import java.awt.peer.ComponentPeer;
import com.ms.awt.peer.ComponentPeerX;

/**
 * A class that produces extended font objects. 
 * @version 	1.15, 5/sept/96
 * @author 		David Silver
 */
public class FontX extends Font
{
    // public static final int PLAIN	= 0;
    // public static final int BOLD	= 1;
	// public static final int ITALIC	= 2;
    // private int pData;
    // private String family;
    // protected String name;
    // protected int style;
    // protected int size;
    
    /**
     * Flag used to indicate the new font should be linked to a privately 
     * embedded font. (FUTURE: This is only nessacary if the font was loaded by 
     * a non-java peice of code. If the font was loaded into the system
     * through Java, the font will be known to be embedded).
     */
	public static final int EMBEDDED = 1;
	
    /**
     * Flag used to indicate that text drawn using the the new font should be stuck out.
     */
	public static final int STRIKEOUT = 2;
	
    /**
     * Flag used to indicate that text drawn using the the new font should be underlined.
     */
	public static final int UNDERLINE = 4;

	/**
	 * Flag used to indicate that the text drawn using the new font should be
	 * drawn using the outline of the glyph. This is only supported by certain
	 * fonts in the system, and only GraphicsX will support this feature. Trying
	 * to draw outline fonts using a Win32Graphics object will result in the glyph
	 * being drawn normally.
	 */
	public static final int OUTLINE = 8;
	
	private static final int FONTXFONT = 0x8000;
	
    private native static int pNewFontEnumeration();
	private native static String pGetFontEnumeratedFamily(int i);
	private native boolean pIsTypeable(int lang);
	private native static FontX pChooseFont(int c, FontX f);
	
	private int flags = FONTXFONT;		// FONTXFONT is used by the native side.
	
	//========================================================================
	
    /**
     * Create an extended font.
     *
     * @param name	the face name of the font.
     * @param style	the style of the font.
     *				For compatibility reasons, this is limited to the same
     *				styles as the fonts in Java.awt.Font.
     * @param size	the size of the font.
     * @return	the newly created object.
     */
	public FontX(String name, int style, int size) 
    {
    	super(name, style, size);
    }

    public FontX(String name, int style, int size, boolean bEmbed) 
    {
    	super(name, style, size);
    	if ( bEmbed )
    		flags |= 1;
    	else
    		flags &= 0xfffe;
    }

    /**
     * Create an extended font.
     *
     * @param name		the	face name of the font.
     * @param style		the style of the font.
     *					For compatibility reasons, this is limited to the same
     *					styles as the fonts in Java.awt.Font.
     * @param size		the size of the font.
     * @param xFlags	extended styles.  To ensure that future versions of 
     *					AWT will not clash with FontX, the extensions to the style
     *				    are put in a separate parameter. The current options are...
     *		EMBEDDED
     *		UNDERLINE
     *		STRIKEOUT
     *		OUTLINE		
     *		SHADOW		(TBD)
     * @return	the newly created object.
     */
    public FontX(String name, int style, int size, int xFlags) 
    {
    	super(name, style, size);
		flags |= xFlags;
    }

	//========================================================================

    /**
     * Check whether a font supports characters in a language. 
     *
     * @param language		The language to be queried.
     * @return	true or false.
     */
	public boolean isTypeable(int language)
	{
		return pIsTypeable(language);
	}

	//========================================================================
    /**
     * Return the extended flags of a FontX object. 
     * @return	the extended styles. See EMBEDDED, STRIKEOUT, UNDERLINE etc.
     */
	public int getFlags()
	{
		return flags;
	}
	
	//========================================================================
/*
 * All the following need modifying for the new class - for tt extensions
 *
    public String getFamily() 
    {
		return super.getFamily();
    }

	//========================================================================
    public String getName() 
    {
		return super.getName();
    }

	//========================================================================
    public int getStyle() 
    {
		return super.getStyle();
    }

	//========================================================================
    public int getSize() 
    {
		return super.getSize();
    }

	//========================================================================
    public boolean isPlain() 
    {
	return 	super.isPlain();
    }

	//========================================================================
    public boolean isBold() 
    {
		return super.isBold();
    }

	//========================================================================
    public boolean isItalic() 
    {
		return super.isItalic();
    }

	//========================================================================
	public int hashCode() 
    {
		return (super.getName()).hashCode() ^ style ^ size;
	}
*/
	//========================================================================
	
    public boolean equals(Object obj) 
    {
    	if( obj instanceof FontX )
    		return super.equals(obj);
    	else if (obj instanceof Font) 
    		return super.equals(obj);
    	else
			return false;
    }

	//========================================================================
	
    public String toString() 
    {
		String	strStyle;

		if (isBold()) 
			strStyle = isItalic() ? "bolditalic" : "bold";
			
		else 
		    strStyle = isItalic() ? "italic" : "plain";

		return getClass().getName() + 
					"[family=" + getFamily() + 
					",name=" + getName() + 
					",style=" + strStyle + 
					",size=" + getSize() + 
					(((flags & EMBEDDED)!=0) ?", Embedded" :"") +
					(((flags & STRIKEOUT)!=0) ?", Strikeout" :"") +
					(((flags & UNDERLINE)!=0) ?", Underline" :"") +
					(((flags & OUTLINE)!=0) ?", Outline" :"") +
					"]";
    }
    
	//========================================================================
	
    /**
     * Return a list of all fonts in the system. 
     *
     * @return	An array of strings each of which holds the face names.
     */
    public static String[] getFontList() 
    {
    	int listSize = pNewFontEnumeration();
    	
    	String list[] = new String[listSize + 6];
    	list[0] = "Dialog";
    	list[1] = "Helvetica";
    	list[2] = "TimesRoman";
		list[3] = "Courier";
    	list[4] = "DialogInput";
    	list[5] = "ZapfDingbats";

    	for(int i=0; i<listSize; i++)
    		list[6+i] = pGetFontEnumeratedFamily(i);
		return list;
    }
	
	//========================================================================

	/**
	 * put up a dialog box to allow the user to choose a font. A default font
	 * will be initialised.
	 *
	 * @param c		The component who will be blocked by owning the dialog box
	 * @return A FontX object or null if the user hits the cancel button
	 */
	public static FontX ChooseFont(Component c)
	{
		return ChooseFont(c, null);
	}
	
	//========================================================================

	/**
	 * put up a dialog box to allow the user to choose a font. The default font
	 * will be initialised to the supplied font.
	 *
	 * @param c		The component who will be blocked by owning the dialog box
	 * @param f		The FontX object to use as the default for the new font
	 * @return A FontX object or null if the user hits the cancel button
	 */
	public static FontX ChooseFont(Component c, FontX f)
	{
		ComponentPeer p = c.getPeer();
		ComponentPeerX peer = (ComponentPeerX)p;
		if( peer == null )
			f = pChooseFont(0, f);
		else
			f = pChooseFont(peer.gethwnd(), f);
		return f;
	}
	//========================================================================
}

