//
//  Copyright (c) 1996, Microsoft Corporation
//
package com.ms.awt;

import sun.awt.win32.*;
import com.ms.awt.FontX;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Point;

/**
 * A class then encapsulates an extended graphics object
 * @version 1.0 3/sept/96
 * @author David Silver
 */
public class GraphicsX extends Win32Graphics 
{
	private final int TTPOLYHEADER = 0x10;
	private GlyphOutline go[] = null; 	// only used if need an outline font
	
	private native byte[] getCharOutline(char ch);
	private native void pDrawBezier(int xOrig, int yOrig, PointFx start, OutlineCurve oc, int nPoints);
	
	/**
	 * Create a graphics object for a print job
	 */
    public GraphicsX(Win32PrintJob pj) 
    {
    	super(pj);
    }
    
	/**
	 * Create a graphics object based on another Graphics Object
	 */
    public GraphicsX(GraphicsX g) 
    {
    	super(g);
    }
    
	/**
	 * Create a graphics object based on another Graphics Object
	 */
    public GraphicsX(Win32Graphics g) 
    {
    	super(g);
    }
    
	/**
	 * Create a graphics object for a component
	 */
    public GraphicsX(MComponentPeer comp) 
    {
    	super(comp);
    }
    
	/**
	 * Create a graphics object for an off-screen image.
	 */
    public GraphicsX(Win32Image image) 
    {
    	super(image);
    }

	/**
	 * retrieve the Win32 display context handle
	 */
	public int gethdc()
	{
		return GetDC();
	}

    /**
     * Sets the font for all subsequent text-drawing operations.
     * @param font the specified font
     * @see Font
     * @see #getFont
     * @see #drawString
     * @see #drawBytes
     * @see #drawChars
     */
    public void setFont(Font font)
    {
    	// first clean out the glyph arrays if the current font is an outline font
    	if( go != null )
    		go = null;
    	super.setFont(font);
    }
    
	private FontX getFontX( int flags )
	{
    	Font f = getFont();
    	if( f instanceof com.ms.awt.FontX )
    	{
    		FontX fx = (FontX)f;
    		if( flags == 0 )
    			return fx;
    		if( (fx.getFlags() & flags) != 0 )
    			return fx;
    	}
    	return null;
	}
	
    /** 
     * Draws the given character array. 
     */
    public void drawChars(char data[], int offset, int length, int x, int y)
    {
    	FontX fx = getFontX(FontX.OUTLINE);
    	if( fx == null )
    		super.drawChars(data, offset, length, x, y);
    	else
    		drawCharsWidth(data, offset, length, x, y);
    }

    /** 
     * Draws the given character array and return the width in
     * pixels. If font is not set then returns -1. 
     */
    public int drawCharsWidth(char data[], int offset, int length, int x, int y)
    {
    	FontX fx = getFontX(FontX.OUTLINE);
    	if( fx == null )
    		return super.drawCharsWidth(data, offset, length, x, y);
    	else
    	{
    		int i;
    		FontMetrics fm = getFontMetrics();
			int ascent = fm.getAscent();

    		for(i=0; i<length; i++)
			{
				GlyphOutline go = getGlyphOutline(data[offset+i]);
				if( go != null)
					drawOutlineChar(x, y+ascent, go);
				else
					super.drawCharsWidth(data, offset+i, 1, x, y);
				x += fm.charWidth(data[offset+i]);
			}
			return x;
    	}
    }

    /** 
     * Draws the given byte array. 
     */
    public void drawBytes(byte data[], int offset, int length, int x, int y)
    {
    	FontX fx = getFontX(FontX.OUTLINE);
    	if( fx == null )
    		super.drawBytes(data, offset, length, x, y);
    	else
    	{
    		super.drawBytes(data, offset, length, x, y);
    	}
    }

    /** 
     * Draws the given string. 
     */
	public void drawString(String str, int x, int y)
	{
		FontX fx = getFontX(FontX.OUTLINE);
		if( fx == null )
			super.drawString(str, x, y);
		else
			drawStringWidth(str, x, y);	// calling drawStringWidth breaks!
	}

    /** 
     * Draws the given string and returns the length of the drawn
     * string in pixels.  If font is not set then returns -1. 
     */
    public int drawStringWidth(String str, int x, int y)
    {
    	FontX fx = getFontX(FontX.OUTLINE);
    	if( fx == null )
    		return super.drawStringWidth(str, x, y);
    	else
    		return super.drawStringWidth(str, x, y);
    }

    /** 
     * Draws the given character array and return the width in
     * pixels. If font is not set then returns -1. 
     */
    public int drawBytesWidth(byte data[], int offset, int length, int x, int y)
    {
    	FontX fx = getFontX(FontX.OUTLINE);
    	if( fx == null )
    		return drawBytesWidth(data, offset, length, x, y);
    	else
    	{
    		return drawBytesWidth(data, offset, length, x, y);
    	}
    }

	/**
	 * get the outline of a character in the currently selected font. If
	 * the font is not an outline font, then null is returned
	 *
	 * @param ch the character whose outline is requested
	 * @return an OutlinePolygon object
	 */
	public GlyphOutline getGlyphOutline(char ch)
	{
		// we only store the first 256
		if(ch >= (char)256)
			return getGlyphOutline1(ch);

		// only allocate the outline array if we need it.
		if( go == null )
			go = new GlyphOutline[256];
			
		// only allocate the outline shape if we need it.
		if( go[ch] == null )
			go[ch] = getGlyphOutline1(ch);

		return go[ch];
	}

	//--------------------------------------------------------------------------
	//--------------------------------------------------------------------------
	private int bPos ;
	
	//
	// an int in Java is 32 bits - get a long from the stream
	//
	int getSizeGlyphInt()
	{
		return 4;
	}

	int getGlyphInt(byte b[])
	{
		int i = (b[bPos] & 0xff) + ((b[bPos+1] & 0xff) * 256) +
			    ((b[bPos+2] & 0xff) + (b[bPos+3] & 0xff) * 256)*65536;
		bPos += getSizeGlyphInt();
		return i;
	}

	//
	// an int in Java is 32 bits - get a short from the stream
	//
	int getGlyphShort(byte b[])
	{
		int i = (b[bPos] & 0xff) + ((b[bPos+1] & 0xff) * 256);
		bPos += 2;
		return i;
	}

	Point getGlyphPoint(byte b[])
	{
		int x = getGlyphInt(b);
		int y = getGlyphInt(b);
		return new Point(x, y);
	}

	float getGlyphFloat(byte b[])
	{
		int value   = getGlyphInt(b);
		float theVal = (float)value/(float)65536.0;
		return theVal;
	}

	int getSizeGlyphFloat(byte b[])
	{
		return getSizeGlyphInt();
	}

	PointFx getGlyphPointFx(byte b[])
	{
		float x;
		float y;

		x = getGlyphFloat(b);
		y = getGlyphFloat(b);
		return new PointFx(x,y);
	}

	int getSizeGlyphPointFx(byte b[])
	{
		return getSizeGlyphFloat(b) * 2;
	}

	private GlyphOutline getGlyphOutline1(char ch)
	{
		int bStart;
		int nBytes;
		byte b[] = getCharOutline(ch);
		if( b == null )
			return null;
			
		int bLen = b.length;
		bPos = 0;

		GlyphOutline go = new GlyphOutline();
		GlyphMetrics gm = new GlyphMetrics();
		gm.blackBoxX   = getGlyphInt(b);
		gm.blackBoxY   = getGlyphInt(b);
		gm.glyphOrigin = getGlyphPoint(b);
		gm.cellIncX    = getGlyphShort(b);
		gm.cellIncY    = getGlyphShort(b);
		go.gm          = gm;

		// now find out how many OutlinePolygons we have
		int nOutlineBytes = bLen - bPos;
		int nPolygons = 0;
		bStart = bPos;
		while(nOutlineBytes != 0)
		{
			nPolygons++;
			nBytes = getGlyphInt(b);
			nOutlineBytes -= nBytes;
			bPos += (nBytes-getSizeGlyphInt());
		}

		// now create the polygons
		//
		go.polygons = new OutlinePolygon[nPolygons];
		for(int iPolygon = 0; iPolygon<nPolygons; iPolygon++)
		{
			bPos             = bStart;
			int nPolygonSize = getGlyphInt(b);
			nBytes           = nPolygonSize - TTPOLYHEADER;

			OutlinePolygon op = new OutlinePolygon(OutlinePolygon.OUTLINE_TRUETYPE);
			go.polygons[iPolygon] = op;
			
			op.type    = (getGlyphInt(b) == 0x18) ?OutlinePolygon.OUTLINE_TRUETYPE 
				                                 :OutlinePolygon.OUTLINE_UNDEFINED;
			op.fxStart = getGlyphPointFx(b);	

			// now need to work out size of arrays.
			//
			int nCurves = 0;
			while(nBytes > 0)
			{
				// move to the end of this curve.
				int type = getGlyphShort(b);
				int cpfx = getGlyphShort(b);
				int i = getSizeGlyphPointFx(b) * cpfx;
				bPos += i;
				nBytes -= (i+4);
				nCurves ++;
			}
			op.curves = new OutlineCurve[nCurves];

			// now go back and create the curves
			bPos = bStart+TTPOLYHEADER;
			for( int iCurve = 0; iCurve<nCurves; iCurve++)
			{
				OutlineCurve oc = new OutlineCurve();
				op.curves[iCurve] = oc;

				int type = getGlyphShort(b);
				oc.type = type;

				if( (type != OutlineCurve.PRIM_LINE) && 
					(type != OutlineCurve.PRIM_QSPLINE) )
				{
					System.out.println("Error in parsing OutlinePolygon!");
					return null;
				}

				int len = getGlyphShort(b);
				oc.points = new PointFx[len];
				for(int iPoint = 0; iPoint < len; iPoint++)
					oc.points[iPoint] = getGlyphPointFx(b);
			}
			bStart += nPolygonSize;
		}
		return go;
	}
	
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////
	//
	//   Formula for the T2 B-Spline:
	//
	//     f(t) = (A-2B+C)*t^2 + (2B-2A)*t + A
	//  
	//   where
	//
	//     t = 0..1
	//
	private void drawT2Curve( PointFx ptlA, PointFx ptlB, PointFx ptlC, 
								int xOrig, int yOrig )
	{
		int x, y, x1, y1;
		double ax, vx, x0, ay, vy, y0, t;

		ax = ptlA.x - 2*ptlB.x + ptlC.x;
		vx = 2*ptlB.x - 2*ptlA.x;
		x0 = ptlA.x;
		
		ay = ptlA.y - 2*ptlB.y + ptlC.y;
		vy = 2*ptlB.y - 2*ptlA.y;
		y0 = ptlA.y;

		x1 = xOrig+(int)(x0);
		y1 = yOrig-(int)(y0);

		for( t = 0.0; t < 1.0; t += (1.0/10.0) )
		{
			x = xOrig + (int)((double)(ax*t*t + vx*t + x0));
			y = yOrig - (int)((double)(ay*t*t + vy*t + y0));
			drawLine(x1, y1, x, y);
			x1 = x;
			y1 = y;
	   }
	}

	/**
	 * draw an outline polygon
	 *
	 * @param xOrig - the horizontal starting point for the shape
	 * @param yOrig - the vertical starting point for the shape
	 * @param op	- the outline polygon
	 */
	public void drawOutlinePolygon(int xOrig, int yOrig, OutlinePolygon op)
	{
		PointFx ptLast = op.fxStart;
		int nCurves    = op.curves.length;
		PointFx ptlC   = new PointFx();		// need this for qspline
		
		for(int iCurve=0; iCurve<nCurves; iCurve++)
		{
			OutlineCurve oc = op.curves[iCurve];
			int nPoints = oc.points.length;

			if(oc.type == OutlineCurve.PRIM_LINE)
			{
				for(int iPoint=0; iPoint<nPoints; iPoint++)
				{
					PointFx pt = oc.points[iPoint];

					drawLine(xOrig+(int)ptLast.x, yOrig-(int)ptLast.y, 
							 xOrig+(int)pt.x,     yOrig-(int)pt.y);
					ptLast = pt;
				}
			}
			else if(oc.type == OutlineCurve.PRIM_QSPLINE)
			{
				if( op.type == OutlinePolygon.OUTLINE_POLYBEZIER )
				{
					pDrawBezier(xOrig, yOrig, ptLast, oc, nPoints);
				}
				else if( op.type == OutlinePolygon.OUTLINE_TRUETYPE )
				{
					PointFx ptlA = ptLast;
					PointFx ptlB;
					PointFx ptlD;

					for(int iPoint=0; iPoint<(nPoints-1); iPoint++)
					{
						ptlB = oc.points[iPoint];

						if(iPoint < (nPoints-2))
						{
							// if not on last spline, compute C
							ptlC.x = (ptlB.x + oc.points[iPoint+1].x) / (float)2.0;
							ptlC.y = (ptlB.y + oc.points[iPoint+1].y) / (float)2.0;
							ptlD = ptlC;
						}
						else
							ptlD = oc.points[iPoint+1];

						drawT2Curve(ptlA, ptlB, ptlD, xOrig, yOrig);
						ptlA = ptlD;
					}
				}
				else
				{
					System.out.println("Error! Unknown polygon type");
					return;
				}
			ptLast = oc.points[nPoints-1];
			}
		}

		if( (op.fxStart.x != ptLast.x) || (op.fxStart.y != ptLast.y) )
			drawLine(xOrig+(int)ptLast.x, yOrig-(int)ptLast.y, 
		    	     xOrig+(int)op.fxStart.x, yOrig-(int)op.fxStart.y);
	}
	
	private void drawOutlineChar(int xOrig, int yOrig, GlyphOutline go)
	{
		int nPolygons = go.polygons.length;

		for(int iPolygon=0; iPolygon<nPolygons; iPolygon++)
			drawOutlinePolygon(xOrig, yOrig, go.polygons[iPolygon]);
	}
	
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////
}
