//
//  Copyright (c) 1996, Microsoft Corporation
//
package com.ms.lang;

import java.util.Date;

/**
 * A class that produces registry access objects.
 *
 * note: the following are NOT implemented....
 *		RegConnectRegistry
 *		RegCreateKeyEx
 *		RegGetKeySecurity
 *  	RegNotifyChangeKeyValue
 *		RegOpenKeyEx
 *		RegQueryMultipleValues
 *		RegQueryValue
 *		RegSaveKey
 *		RegSetKeySecurity
 *		RegSetValue
 * @version 	1.15, 5/sept/96
 * @author 		David Silver
 */
public class RegKey
{
	private final static int HKEY_CLASSES_ROOT = 0x80000000;
	private final static int HKEY_CURRENT_USER = 0x80000001;
	private final static int HKEY_LOCAL_MACHINE = 0x80000002;
	private final static int HKEY_USERS = 0x80000003;

	/**
	 * ID signifying a key should be opened for READ access
	 */
	public final static int KEYOPEN_READ = 1;
	
	/**
	 * ID signifying a key should be opened for WRITE access
	 */
	public final static int KEYOPEN_WRITE = 2;

	/**
	 * ID signifying a key should be opened for ALL access
	 */
	public final static int KEYOPEN_ALL = 3;
	
	/**
	 * ID signifying a key should be created, not opened
	 */
	public final static int KEYOPEN_CREATE = 4;
	
	/**
	 * ID signifying that the HKEY_CLASSES_ROOT key should be returned
	 */
	public final static int CLASSES_ROOT = 1;
	
	/**
	 * ID signifying that the HKEY_LOCAL_MACHINE key should be returned
	 */
	public final static int LOCALMACHINE_ROOT = 2;
	
	/**
	 * ID signifying that the HKEY_CURRENT_USER key should be returned
	 */
	public final static int USER_ROOT = 3;
	/**
	 * ID signifying that the HKEY_USERS key should be returned
	 */
	public final static int USERS_ROOT = 4;
	
	private boolean keyOpen = false;
	private int regKey = 0;			// this is the key!!
	private boolean fullKey = false;
	
	private native void pRegCreateKey(RegKey r, String s);
	private native void pRegLoadKey(String subKey, String fileName);
	private native void pRegDeleteKey(String s);

	private native void pRegOpenBaseKey(int l);
	private native void pRegOpenKey(RegKey r, String s, int access);

	private native void pRegCloseKey();
	private native RegKeyEnumValue pRegEnumValue(int i);
	private native String pRegEnumKey(int i);
	private native void pRegFlushKey();
	private native String pRegQueryStringValue(String name);
	private native byte[] pRegQueryBinaryValue(String name);
	private native int pRegQueryIntValue(String name);
	private native int pRegQueryInfoKey(RegQueryInfo rqi);
	private native void pRegSetStringValue(String s, String v);
	private native void pRegSetIntValue(String s, int v);
	private native void pRegSetBinaryValue(String s, byte v[]);
	private native void pRegDeleteValue(String val);
	private native void pRegRestoreKey(String filename, boolean vol);
	private native void pRegUnLoadKey(String subKey);
	private native void pRegReplaceKey(String subKey, String newFile, String oldFile);

	///////////////////////////////////////////////////////////////////////////////
	// Private entry point used to create the base keys

	private final static RegKey hkeyClassesRoot = new RegKey(HKEY_CLASSES_ROOT);
	private final static RegKey hkeyCurrentUser = new RegKey(HKEY_CURRENT_USER);
	private final static RegKey hkeyLocalMachine = new RegKey(HKEY_LOCAL_MACHINE);
	private final static RegKey hkeyUsers = new RegKey(HKEY_USERS);
	
	private RegKey(long hkeyRoot)
	{
		pRegOpenBaseKey((int)hkeyRoot);
	}
	
	///////////////////////////////////////////////////////////////////////////////
	/**
	 * Get a top level key
	 */
	public final static RegKey getRootKey(int id)
	{
		SecurityManager security = System.getSecurityManager();
		if (security != null) 
   			security.checkRegistry();

   		switch (id)
   		{
   			case CLASSES_ROOT :		return hkeyClassesRoot;
   			case LOCALMACHINE_ROOT:	return hkeyLocalMachine;
   			case USER_ROOT:			return hkeyCurrentUser;
   			case USERS_ROOT:		return hkeyUsers;
   			default:				
   				// throw an exception here!
   				return null;
  		}
	}

	///////////////////////////////////////////////////////////////////////////////
	/*
	 * finalize - clean up the key.
	 */
	public void finalize()
	{
		if( keyOpen )
			close();
	}

	///////////////////////////////////////////////////////////////////////////////
	/**
	 * Create the specified key. If the key already exists in the registry, the 
	 * function opens it.
	 *
	 * @param rk	Identifies a currently open key or any of the following 
	 *				predefined reserved handle values: 
	 *					null
	 *					HKEY_CLASSES_ROOT
	 *					HKEY_CURRENT_USER
	 *					HKEY_LOCAL_MACHINE
	 *					HKEY_USERS
	 *				The key created is a subkey of the key identified by the rk 
	 *				parameter. 
	 * 				If the applet is running in a browser, or other protected
	 *				environment, then this parameter MUST be null.  This 
	 *				specifies that the key will be created in a sandboxed 
	 *				section of the registry.
	 * @param subKey 	Points to a String specifying the name of a subkey 
	 *					that this function opens or creates. The subkey specified 
	 *					must be a subkey of the key identified by the rk parameter. 
	 *					This subkey must not begin with the backslash character ('\'). 
	 *					This parameter cannot be null. 
	 * @param create	If this is true, if the key does not exist it will be created
	 * @return	the newly created registry key.
     * @exception RegKeyException If the key can not be created.
	 */
	public RegKey(RegKey rkParent, String subKey, int access)
	{
		// check to see if the applet is allowed to open files. If so, then
		// its allowed to use the registry as well.
		SecurityManager security = System.getSecurityManager();
		if (security != null) 
   			security.checkRegistry();

//     * @exception RegKeyException If the parent key is not accessible.

//		if(rkParent != null)
//		{
// for now, no applet can open the registry
//			if( rkParent.fullKey )
//			{
//				// check to see if the applet is allowed to open files. If so, then
//				// its allowed to use the registry as well.
//				SecurityManager security = System.getSecurityManager();
//				if (security != null) 
//	    			security.checkFileDialog();
//				fullKey = true;
//			}
//		}
		
		if (access == KEYOPEN_CREATE)
			pRegCreateKey(rkParent, subKey);
		else
			pRegOpenKey(rkParent, subKey, access);
	}

	/**
	 * close should be used when a key has been finished with.  As the system registry keys 
	 * need to be written out to disk, the key should be closed when it is no longer needed.
	 * After this, the key should not be used again.
	 */
	public void close()
	{
		pRegCloseKey();
		keyOpen = false;
	}

	/**
	 * remove a subkey from the key.
	 *
	 * @param sub	the section to prune.
	 * @exception	IOException if unsuccessful.
	 */
	public void deleteSubKey(String sub)
	{
		pRegDeleteKey(sub);
  	}
  	
	/**
	 * delete a value from the key
	 * @param val	the name of the value to be deleted
	 */
	 public void deleteValue(String val)
	 {
	 	pRegDeleteValue(val);
	 }

	/**
	 * enumerate subvalues of the key.
	 *
	 * @param idx	index of the enumeration.
	 * @return subkey	name of the subkey.
	 */
	public RegKeyEnumValue enumValue(int idx)
	{
		return pRegEnumValue(idx);	// returns a RegKeyEnumValue.
	}

	/**
	 * enumerate subkeys of the key.
	 */
	public String enumKey(int idx)
	{
		return pRegEnumKey(idx);
	}
	
	/**
	 * flush the key to the registry on disk.
	 */
	public void flush()
	{
		pRegFlushKey();
	}

	/**
	 * Load a new hive into the registry
	 * @param subKey	name of the subkey to be created to contain the new hive
	 * @param hive		name of a file containing the registry information
	 */
	 public void loadKey(String subKey, String fileName)
	 {	
	 	pRegLoadKey(subKey, fileName);
	 }

	/**
	 * Query a String value from the registry. The name is a sub key of the current key.
	 * @param name 	subKey containing the value wanted.
	 * @return	String value contained in the subKey.
	 * 
	 */
	 public String getStringValue(String name)
	 {
	 	return pRegQueryStringValue(/*REG_SZ,*/ name);
	 }
	 
	/**
	 * Query a byte array from the registry. The name is a sub key of the 
	 * current key.
	 * @param name 	subKey containing the value wanted.
	 * @return	byte array contained in the subKey.
	 * 
	 */
	 public byte[] getBinaryValue(String name)
	 {
	 	return pRegQueryBinaryValue(/*REG_BINARY,*/ name);
	 }
	 
	/**
     * Query an integer value from the registry. The name is a sub key of the current key.
	 * @param name 	subKey containing the value wanted.
	 * @return	Int value contained in the subKey.
	 * 
	 */
     public int getIntValue(String name)
	 {
	 	return pRegQueryIntValue(/*REG_DWORD,*/ name);
	 }

	 /**
	  * Replaces the file backing a key and all its subkeys with another file, so 
	  * that when the system is next started, the key and subkeys will have the 
	  * values stored in the new file. 
	  * @param subKey
	  * @param newFile
	  * @param oldFile
	  */
  	public void replace(String subKey, String newFile, String oldFile)
  	{
  		pRegReplaceKey(subKey, newFile, oldFile);
  	}

  	/**
  	 * restore a subkey from a file
  	 * @param file	the file name
  	 * @param volatile	if the hive is volatile
  	 */
	public void restore( String filename, boolean vol )  	 
	{
		pRegRestoreKey(filename, vol);
	}

	/**
	 * set a String value into a subkey.
	 * @param subKey	the subkey to hold the value
	 * @param val		the string to write out.
	 */
	public void setValue(String subKey, String val)
	{	
		pRegSetStringValue(subKey, val);
	}
	
	/**
	 * set a byte array into a subkey.
	 * @param subKey	the subkey to hold the value
	 * @param val		the array to write out.
	 */
	public void setValue(String subKey, byte val[])
	{	
		pRegSetBinaryValue(subKey, val);
	}
	
	/**
	 * set an int value into a subkey.
	 * @param subKey	the subkey to hold the value
	 * @param val		the int to write out.
	 */
	public void setValue(String subKey, int val)
	{	
		pRegSetIntValue(subKey, val);
	}

	/**
	 * unloads a hive from the registry.
	 * @param subKey the subkey to unload.
	 */
	public void unload(String subKey)
	{
		pRegUnLoadKey(subKey);
	}

	/**
	 * Retreive information about the key.
	 * @return A RegQueryInfo object which contains the information
	 */
	public RegQueryInfo queryInfo()
	{
		RegQueryInfo rqi = new RegQueryInfo();
		
		/*
	     * fatDate is FatDate in the low word, fatTime in the high word
         *
         * Some platforms do not support timestamps of registry keys, so the
         * native API will return zero.
		 * 
		 * Bits		Contents
		 * 0-4		Day of the month (1-31)
		 * 5-8		Month (1 = January, 2 = February, etc.)
		 * 9-15		Year offset from 1980 (add 1980 to get actual year)
		 * 16-20	Second divided by 2
		 * 21-26	Minute (0-59)
		 * 27-31	Hour (0-23 on a 24-hour clock)
		 */
		int fatDate = pRegQueryInfoKey(rqi);

        if (fatDate != 0)
        {
            int day    = fatDate & 0x1f;                   fatDate >>= 5;
            int month  = (fatDate & 0x0f) - 1;             fatDate >>= 4;
            int year   = (fatDate & 0x7f) + 1980 - 1900;   fatDate >>= 7;
            int second = (fatDate & 0x1f) * 2;             fatDate >>= 5;
            int mins   = fatDate & 0x3f;                   fatDate >>= 6;
            int hours  = fatDate & 0x1f;   

            rqi.lastWriteTime = new Date(year, month, day, hours, mins);
        }
        else
            rqi.lastWriteTime = null;

		return rqi;
	}
	
}
